import Module from 'module';
import { dirname, resolve, relative, basename } from 'pathe';
import chalk from 'chalk';
import consola from 'consola';
import defu from 'defu';
import { createHooks } from 'hookable';
import prettyBytes from 'pretty-bytes';
import mkdirp from 'mkdirp';
import fsp, { writeFile } from 'fs/promises';
import { promisify } from 'util';
import _rimraf from 'rimraf';
import jiti from 'jiti';
import { rollup } from 'rollup';
import commonjs from '@rollup/plugin-commonjs';
import { nodeResolve } from '@rollup/plugin-node-resolve';
import alias from '@rollup/plugin-alias';
import _esbuild from 'rollup-plugin-esbuild';
import dts from 'rollup-plugin-dts';
import rollupJSONPlugin from '@rollup/plugin-json';
import { createFilter } from '@rollup/pluginutils';
import { findStaticImports } from 'mlly';
import MagicString from 'magic-string';
import { resolveSchema, generateMarkdown, generateTypes } from 'untyped';
import untypedPlugin from 'untyped/loader/babel';
import { pascalCase } from 'scule';
import { mkdist } from 'mkdist';

async function ensuredir(path) {
  await mkdirp(dirname(path));
}
async function symlink(from, to, force = true) {
  await ensuredir(to);
  if (force) {
    await fsp.unlink(to).catch(() => {
    });
  }
  await fsp.symlink(from, to, "junction");
}
function dumpObject(obj) {
  return "{ " + Object.keys(obj).map((key) => `${key}: ${JSON.stringify(obj[key])}`).join(", ") + " }";
}
function getpkg(id = "") {
  const s = id.split("/");
  return s[0][0] === "@" ? `${s[0]}/${s[1]}` : s[0];
}
const rimraf = promisify(_rimraf);
async function rmdir(dir) {
  await fsp.unlink(dir).catch(() => {
  });
  await rimraf(dir);
}
function tryRequire(id, rootDir = process.cwd()) {
  const _require = jiti(rootDir, { interopDefault: true });
  try {
    return _require(id);
  } catch (err) {
    if (err.code !== "MODULE_NOT_FOUND") {
      console.error(`Error trying import ${id} from ${rootDir}`, err);
    }
    return {};
  }
}

function validateDependencies(ctx) {
  const usedDependencies = /* @__PURE__ */ new Set();
  const unusedDependencies = new Set(Object.keys(ctx.pkg.dependencies || {}));
  const implicitDependnecies = /* @__PURE__ */ new Set();
  for (const id of ctx.usedImports) {
    unusedDependencies.delete(id);
    usedDependencies.add(id);
  }
  if (Array.isArray(ctx.options.dependencies)) {
    for (const id of ctx.options.dependencies) {
      unusedDependencies.delete(id);
    }
  }
  for (const id of usedDependencies) {
    if (!ctx.options.externals.includes(id) && !id.startsWith("chunks/") && !ctx.options.dependencies.includes(getpkg(id))) {
      implicitDependnecies.add(id);
    }
  }
  if (unusedDependencies.size) {
    consola.warn("Potential unused dependencies found:", Array.from(unusedDependencies).map((id) => chalk.cyan(id)).join(", "));
  }
  if (implicitDependnecies.size && !ctx.options.rollup.inlineDependencies) {
    consola.warn("Potential implicit dependencies found:", Array.from(implicitDependnecies).map((id) => chalk.cyan(id)).join(", "));
  }
}

const EXPORT_DEFAULT = "export default ";
function JSONPlugin(options) {
  const plugin = rollupJSONPlugin(options);
  return {
    ...plugin,
    name: "unbuild-json",
    transform(code, id) {
      const res = plugin.transform.call(this, code, id);
      if (res && typeof res !== "string" && "code" in res && res.code && res.code.startsWith(EXPORT_DEFAULT)) {
        res.code = res.code.replace(EXPORT_DEFAULT, "module.exports = ");
      }
      return res;
    }
  };
}

const defaults = {
  include: [/\.(md|txt|css|htm|html)$/],
  exclude: []
};
function rawPlugin(opts = {}) {
  opts = { ...opts, ...defaults };
  const filter = createFilter(opts.include, opts.exclude);
  return {
    name: "unbuild-raw",
    transform(code, id) {
      if (filter(id)) {
        return {
          code: `export default ${JSON.stringify(code)}`,
          map: null
        };
      }
    }
  };
}

function cjsPlugin(_opts) {
  return {
    name: "unbuild-cjs",
    renderChunk(code, _chunk, opts) {
      if (opts.format === "es") {
        return CJSToESM(code);
      }
      return null;
    }
  };
}
const CJSyntaxRe = /__filename|__dirname|require\(|require\.resolve\(/;
const CJSShim = `

// -- Unbuild CommonJS Shims --
import __cjs_url__ from 'url';
import __cjs_path__ from 'path';
import __cjs_mod__ from 'module';
const __filename = __cjs_url__.fileURLToPath(import.meta.url);
const __dirname = __cjs_path__.dirname(__filename);
const require = __cjs_mod__.createRequire(import.meta.url);
`;
function CJSToESM(code) {
  if (code.includes(CJSShim) || !CJSyntaxRe.test(code)) {
    return null;
  }
  const lastESMImport = findStaticImports(code).pop();
  const indexToAppend = lastESMImport ? lastESMImport.end : 0;
  const s = new MagicString(code);
  s.appendRight(indexToAppend, CJSShim);
  return {
    code: s.toString(),
    map: s.generateMap()
  };
}

const esbuild = _esbuild.default || _esbuild;
async function rollupBuild(ctx) {
  if (ctx.options.stub) {
    for (const entry of ctx.options.entries.filter((entry2) => entry2.builder === "rollup")) {
      const output = resolve(ctx.options.rootDir, ctx.options.outDir, entry.name);
      if (ctx.options.rollup.emitCJS) {
        await writeFile(output + ".cjs", `module.exports = require('jiti')(null, { interopDefault: true })('${entry.input}')`);
      }
      await writeFile(output + ".mjs", `import jiti from 'jiti';
export default jiti(null, { interopDefault: true })('${entry.input}');`);
      await writeFile(output + ".d.ts", `export * from '${entry.input}';
export { default } from '${entry.input}';`);
    }
    await ctx.hooks.callHook("rollup:done", ctx);
    return;
  }
  const rollupOptions = getRollupOptions(ctx);
  await ctx.hooks.callHook("rollup:options", ctx, rollupOptions);
  if (!Object.keys(rollupOptions.input).length) {
    return;
  }
  const buildResult = await rollup(rollupOptions);
  await ctx.hooks.callHook("rollup:build", ctx, buildResult);
  const allOutputOptions = rollupOptions.output;
  for (const outputOptions of allOutputOptions) {
    const { output } = await buildResult.write(outputOptions);
    for (const entry of output.filter((e) => e.type === "chunk")) {
      for (const id of entry.imports) {
        ctx.usedImports.add(id);
      }
      if (entry.isEntry) {
        ctx.buildEntries.push({
          path: relative(ctx.options.rootDir, resolve(outputOptions.dir, entry.fileName)),
          bytes: entry.code.length * 4,
          exports: entry.exports
        });
      }
    }
  }
  if (ctx.options.declaration) {
    rollupOptions.plugins = rollupOptions.plugins || [];
    rollupOptions.plugins.push(dts({ respectExternal: true }));
    await ctx.hooks.callHook("rollup:dts:options", ctx, rollupOptions);
    const typesBuild = await rollup(rollupOptions);
    await ctx.hooks.callHook("rollup:dts:build", ctx, typesBuild);
    await typesBuild.write({
      dir: resolve(ctx.options.rootDir, ctx.options.outDir),
      format: "esm"
    });
  }
  await ctx.hooks.callHook("rollup:done", ctx);
}
function getRollupOptions(ctx) {
  const extensions = [".ts", ".mjs", ".cjs", ".js", ".json"];
  return {
    context: ctx.options.rootDir,
    input: Object.fromEntries(ctx.options.entries.filter((entry) => entry.builder === "rollup").map((entry) => [entry.name, resolve(ctx.options.rootDir, entry.input)])),
    output: [
      ctx.options.rollup.emitCJS && {
        dir: resolve(ctx.options.rootDir, ctx.options.outDir),
        entryFileNames: "[name].cjs",
        chunkFileNames: "chunks/[name].cjs",
        format: "cjs",
        exports: "auto",
        preferConst: true,
        externalLiveBindings: false,
        freeze: false
      },
      {
        dir: resolve(ctx.options.rootDir, ctx.options.outDir),
        entryFileNames: "[name].mjs",
        chunkFileNames: "chunks/[name].mjs",
        format: "esm",
        exports: "auto",
        preferConst: true,
        externalLiveBindings: false,
        freeze: false
      }
    ].filter(Boolean),
    external(id) {
      const pkg = getpkg(id);
      const isExplicitExternal = ctx.options.externals.includes(pkg);
      if (isExplicitExternal) {
        return true;
      }
      if (ctx.options.rollup.inlineDependencies || id[0] === "." || id[0] === "/" || id.match(/src[\\/]/) || id.startsWith(ctx.pkg.name)) {
        return false;
      }
      if (!isExplicitExternal) {
        consola.warn(`Inlining implicit external ${id}`);
      }
      return isExplicitExternal;
    },
    onwarn(warning, rollupWarn) {
      if (!warning.code || !["CIRCULAR_DEPENDENCY"].includes(warning.code)) {
        rollupWarn(warning);
      }
    },
    plugins: [
      alias({
        [ctx.pkg.name]: ctx.options.rootDir
      }),
      nodeResolve({
        extensions,
        preferBuiltins: true
      }),
      JSONPlugin({
        preferConst: true
      }),
      esbuild({
        target: "es2020"
      }),
      commonjs({
        extensions,
        ignoreTryCatch: true
      }),
      { renderDynamicImport() {
        return { left: "import(", right: ")" };
      } },
      ctx.options.rollup.cjsBridge && cjsPlugin(),
      rawPlugin()
    ].filter(Boolean)
  };
}

async function typesBuild(ctx) {
  const entries = ctx.options.entries.filter((entry) => entry.builder === "untyped");
  await ctx.hooks.callHook("untyped:entries", ctx, entries);
  for (const entry of entries) {
    const options = {
      jiti: {
        interopDefault: true,
        transformOptions: {
          babel: {
            plugins: [
              untypedPlugin
            ]
          }
        }
      }
    };
    await ctx.hooks.callHook("untyped:entry:options", ctx, entry, options);
    const _require = jiti(ctx.options.rootDir, options.jiti);
    const distDir = entry.outDir;
    const srcConfig = _require(resolve(ctx.options.rootDir, entry.input));
    const defaults = entry.defaults || {};
    const schema = resolveSchema(srcConfig, defaults);
    await ctx.hooks.callHook("untyped:entry:schema", ctx, entry, schema);
    const outputs = {
      markdown: {
        fileName: resolve(distDir, `${entry.name}.md`),
        contents: generateMarkdown(schema)
      },
      schema: {
        fileName: `${entry.name}.schema.json`,
        contents: JSON.stringify(schema, null, 2)
      },
      defaults: {
        fileName: `${entry.name}.defaults.json`,
        contents: JSON.stringify(defaults, null, 2)
      },
      declaration: entry.declaration ? {
        fileName: `${entry.name}.d.ts`,
        contents: generateTypes(schema, pascalCase(entry.name + "-schema"))
      } : void 0
    };
    await ctx.hooks.callHook("untyped:entry:outputs", ctx, entry, outputs);
    for (const output of Object.values(outputs)) {
      await writeFile(resolve(distDir, output.fileName), output.contents, "utf8");
    }
  }
  await ctx.hooks.callHook("untyped:done", ctx);
}

async function mkdistBuild(ctx) {
  const entries = ctx.options.entries.filter((e) => e.builder === "mkdist");
  await ctx.hooks.callHook("mkdist:entries", ctx, entries);
  for (const entry of entries) {
    const distDir = entry.outDir;
    if (ctx.options.stub) {
      await rmdir(distDir);
      await symlink(entry.input, distDir);
    } else {
      const mkdistOptions = {
        rootDir: ctx.options.rootDir,
        srcDir: entry.input,
        distDir,
        format: entry.format,
        cleanDist: false,
        declaration: entry.declaration,
        ext: entry.ext
      };
      await ctx.hooks.callHook("mkdist:entry:options", ctx, entry, mkdistOptions);
      const output = await mkdist(mkdistOptions);
      ctx.buildEntries.push({
        path: distDir,
        chunks: [`${output.writtenFiles.length} files`]
      });
      await ctx.hooks.callHook("mkdist:entry:build", ctx, entry, output);
    }
  }
  await ctx.hooks.callHook("mkdist:done", ctx);
}

async function build(rootDir, stub, inputConfig = {}) {
  rootDir = resolve(process.cwd(), rootDir || ".");
  const buildConfig = tryRequire("./build.config", rootDir) || {};
  const pkg = tryRequire("./package.json", rootDir);
  let preset = buildConfig.preset || pkg.unbuild?.preset || pkg.build?.preset || inputConfig.preset || {};
  if (typeof preset === "string") {
    preset = tryRequire(preset, rootDir);
  }
  const options = defu(buildConfig, pkg.unbuild || pkg.build, inputConfig, preset, {
    rootDir,
    entries: [],
    clean: true,
    declaration: false,
    outDir: "dist",
    stub,
    externals: [...Module.builtinModules],
    dependencies: [],
    devDependencies: [],
    peerDependencies: [],
    rollup: {
      emitCJS: true,
      cjsBridge: false,
      inlineDependencies: false
    }
  });
  options.outDir = resolve(options.rootDir, options.outDir);
  const ctx = {
    options,
    pkg,
    buildEntries: [],
    usedImports: /* @__PURE__ */ new Set(),
    hooks: createHooks()
  };
  if (preset.hooks) {
    ctx.hooks.addHooks(preset.hooks);
  }
  if (inputConfig.hooks) {
    ctx.hooks.addHooks(inputConfig.hooks);
  }
  if (buildConfig.hooks) {
    ctx.hooks.addHooks(buildConfig.hooks);
  }
  options.entries = options.entries.map((entry) => typeof entry === "string" ? { input: entry } : entry);
  for (const entry of options.entries) {
    if (typeof entry.name !== "string") {
      entry.name = basename(entry.input);
    }
    if (!entry.input) {
      throw new Error("Missing entry input: " + dumpObject(entry));
    }
    if (!entry.builder) {
      entry.builder = entry.input.endsWith("/") ? "mkdist" : "rollup";
    }
    if (options.declaration !== void 0 && entry.declaration === void 0) {
      entry.declaration = options.declaration;
    }
    entry.input = resolve(options.rootDir, entry.input);
    entry.outDir = resolve(options.rootDir, entry.outDir || options.outDir);
  }
  options.dependencies = Object.keys(pkg.dependencies || {});
  options.peerDependencies = Object.keys(pkg.dependencies || {});
  options.devDependencies = Object.keys(pkg.devDependencies || {});
  options.externals.push(...options.dependencies, ...options.peerDependencies);
  await ctx.hooks.callHook("build:before", ctx);
  consola.info(chalk.cyan(`${options.stub ? "Stubbing" : "Building"} ${pkg.name}`));
  if (process.env.DEBUG) {
    consola.info(`${chalk.bold("Root dir:")} ${options.rootDir}
  ${chalk.bold("Entries:")}
  ${options.entries.map((entry) => "  " + dumpObject(entry)).join("\n  ")}
`);
  }
  if (options.clean) {
    for (const dir of new Set(options.entries.map((e) => e.outDir).sort())) {
      await rmdir(dir);
      await mkdirp(dir);
    }
  }
  await typesBuild(ctx);
  await mkdistBuild(ctx);
  await rollupBuild(ctx);
  if (options.stub) {
    await ctx.hooks.callHook("build:done", ctx);
    return;
  }
  consola.success(chalk.green("Build succeed for " + pkg.name));
  for (const entry of ctx.buildEntries) {
    consola.log(`  ${chalk.bold(entry.path)} (` + [
      entry.bytes && `size: ${chalk.cyan(prettyBytes(entry.bytes))}`,
      entry.exports && `exports: ${chalk.gray(entry.exports.join(", "))}`,
      entry.chunks && `chunks: ${chalk.gray(entry.chunks.join(", "))}`
    ].filter(Boolean).join(", ") + ")");
  }
  validateDependencies(ctx);
  await ctx.hooks.callHook("build:done", ctx);
  consola.log("");
}

export { build as b };
